import decimal
import json
from datetime import datetime

import pandas as pd
import pytest

from opcua_tools.ua_data_types import (
    NodeIdType,
    UABoolean,
    UAByte,
    UAByteString,
    UADateTime,
    UADouble,
    UAEngineeringUnits,
    UAEUInformation,
    UAEURange,
    UAExtensionObject,
    UAFloat,
    UAInt16,
    UAInt32,
    UAInt64,
    UALocalizedText,
    UANodeId,
    UAQualifiedName,
    UARange,
    UASByte,
    UAString,
    UAStructure,
    UAUInt16,
    UAUInt32,
    UAUInt64,
    UAVariant,
    UAXMLElement,
    VariantType,
)


def test_ua_integer_creation_with_incorrect_type():
    with pytest.raises(TypeError):
        UAInt16(value="foo")
    with pytest.raises(TypeError):
        UAUInt16(value="foo")


def test_ua_unsigned_integer_creation_with_negative_integer():
    with pytest.raises(ValueError):
        UAUInt16(value=-42)
    with pytest.raises(ValueError):
        UAUInt32(value=-43)
    with pytest.raises(ValueError):
        UAUInt64(value=-1)


def test_ua_unsigned_integer_creation_with_float_point_number():
    with pytest.raises(TypeError):
        UAUInt16(float(42.2))
    with pytest.raises(TypeError):
        UAUInt16(decimal(42.2))
    with pytest.raises(TypeError):
        UAUInt32(float(42.2))
    with pytest.raises(TypeError):
        UAUInt32(decimal(42.2))
    with pytest.raises(TypeError):
        UAUInt64(float(42.2))
    with pytest.raises(TypeError):
        UAUInt64(decimal(42.2))


def test_ua_signed_integer_creation_with_float_point_number():
    with pytest.raises(TypeError):
        UAInt16(float(42.2))
    with pytest.raises(TypeError):
        UAInt16(decimal(42.2))
    with pytest.raises(TypeError):
        UAInt32(float(42.2))
    with pytest.raises(TypeError):
        UAInt32(decimal(42.2))
    with pytest.raises(TypeError):
        UAInt64(float(42.2))
    with pytest.raises(TypeError):
        UAInt64(decimal(42.2))


def test_ua_unsigned_integer_creation_with_none_value():
    ua_uint16 = UAUInt16(None)
    assert ua_uint16.value is None
    ua_uint32 = UAUInt32(None)
    assert ua_uint32.value is None
    ua_uint64 = UAUInt64(None)
    assert ua_uint64.value is None


def test_ua_signed_integer_creation_with_none_value():
    ua_int16 = UAInt16(None)
    assert ua_int16.value is None
    ua_int32 = UAInt32(None)
    assert ua_int32.value is None
    ua_int64 = UAInt64(None)
    assert ua_int64.value is None


def test_ua_integer_xml_encode_with_value():
    ua_int16 = UAInt16(value=42)
    ua_uint16 = UAUInt16(value=42)
    ua_int32 = UAInt32(value=42)
    ua_uint32 = UAUInt32(value=42)
    ua_int64 = UAInt64(value=42)
    ua_uint64 = UAUInt64(value=42)
    assert ua_int16.xml_encode(include_xmlns=False) == "<Int16>42</Int16>"
    assert ua_uint16.xml_encode(include_xmlns=False) == "<UInt16>42</UInt16>"
    assert ua_int32.xml_encode(include_xmlns=False) == "<Int32>42</Int32>"
    assert ua_uint32.xml_encode(include_xmlns=False) == "<UInt32>42</UInt32>"
    assert ua_int64.xml_encode(include_xmlns=False) == "<Int64>42</Int64>"
    assert ua_uint64.xml_encode(include_xmlns=False) == "<UInt64>42</UInt64>"

    assert (
        ua_int16.xml_encode(include_xmlns=True)
        == '<Int16 xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd">42</Int16>'
    )
    assert (
        ua_uint16.xml_encode(include_xmlns=True)
        == '<UInt16 xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd">42</UInt16>'
    )
    assert (
        ua_int32.xml_encode(include_xmlns=True)
        == '<Int32 xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd">42</Int32>'
    )
    assert (
        ua_uint32.xml_encode(include_xmlns=True)
        == '<UInt32 xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd">42</UInt32>'
    )
    assert (
        ua_int64.xml_encode(include_xmlns=True)
        == '<Int64 xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd">42</Int64>'
    )
    assert (
        ua_uint64.xml_encode(include_xmlns=True)
        == '<UInt64 xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd">42</UInt64>'
    )


def test_ua_integer_json_encode_with_value():
    ua_int16 = UAInt16(value=42)
    ua_uint16 = UAUInt16(value=42)
    ua_int32 = UAInt32(value=42)
    ua_uint32 = UAUInt32(value=42)
    ua_int64 = UAInt64(value=42)
    ua_uint64 = UAUInt64(value=42)
    assert ua_int16.json_encode() == "42"
    assert ua_uint16.json_encode() == "42"
    assert ua_int32.json_encode() == "42"
    assert ua_uint32.json_encode() == "42"
    assert ua_int64.json_encode() == '"42.0"'
    assert ua_uint64.json_encode() == '"42.0"'


def test_ua_integer_json_encode_with_none_value():
    ua_int16 = UAInt16(None)
    ua_uint16 = UAUInt16(None)
    ua_int32 = UAInt32(None)
    ua_uint32 = UAUInt32(None)
    ua_int64 = UAInt64(None)
    ua_uint64 = UAUInt64(None)
    assert ua_int16.json_encode() == None
    assert ua_uint16.json_encode() == None
    assert ua_int32.json_encode() == None
    assert ua_uint32.json_encode() == None
    assert ua_int64.json_encode() == None
    assert ua_uint64.json_encode() == None


def test_ua_floating_point_creation_with_incorrect_type():
    with pytest.raises(TypeError):
        UAFloat(value="foo")
    with pytest.raises(TypeError):
        UADouble(value="foo")


def test_ua_floating_point_creation_casting_from_integer():
    ua_float = UAFloat(value=42)
    assert ua_float.value == float(42)
    ua_double = UADouble(value=42)
    assert ua_double.value == float(42)


def test_ua_floating_point_creation_with_none_value():
    ua_float = UAFloat(None)
    assert isinstance(ua_float.value, pd._libs.missing.NAType)
    ua_double = UADouble(None)
    assert isinstance(ua_double.value, pd._libs.missing.NAType)


def test_ua_floating_point_xml_encode_with_value():
    ua_float = UAFloat(value=42.0)
    ua_double = UADouble(value=42.0)
    assert ua_float.xml_encode(include_xmlns=False) == "<Float>42.0</Float>"
    assert ua_double.xml_encode(include_xmlns=False) == "<Double>42.0</Double>"

    assert (
        ua_float.xml_encode(include_xmlns=True)
        == '<Float xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd">42.0</Float>'
    )
    assert (
        ua_double.xml_encode(include_xmlns=True)
        == '<Double xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd">42.0</Double>'
    )


def test_ua_floating_point_json_encode_with_value():
    ua_float = UAFloat(value=42.0)
    ua_double = UADouble(value=42.0)
    assert ua_float.json_encode() == "42.0"
    assert ua_double.json_encode() == "42.0"

    ua_float = UAFloat(value=float("inf"))
    ua_double = UADouble(value=float("inf"))
    assert ua_float.json_encode() == '"Infinity"'
    assert ua_double.json_encode() == '"Infinity"'

    ua_float = UAFloat(value=float("-inf"))
    ua_double = UADouble(value=float("-inf"))
    assert ua_float.json_encode() == '"-Infinity"'
    assert ua_double.json_encode() == '"-Infinity"'

    ua_float = UAFloat(value=float("nan"))
    ua_double = UADouble(value=float("nan"))
    assert ua_float.json_encode() == '"NaN"'
    assert ua_double.json_encode() == '"NaN"'


def test_ua_floating_point_json_encode_with_none_value():
    ua_float = UAFloat(None)
    ua_double = UADouble(None)
    assert ua_float.json_encode() == None
    assert ua_double.json_encode() == None


def test_ua_string_creation_with_incorrect_type():
    with pytest.raises(TypeError):
        UAString(value=42)


def test_ua_string_creation_with_none_value():
    ua_string = UAString(None)
    assert isinstance(ua_string.value, pd._libs.missing.NAType)


def test_ua_string_should_be_pd_na_if_empty_one_given():
    ua_string = UAString("")

    assert isinstance(ua_string.value, pd._libs.missing.NAType)


def test_ua_string_xml_encode():
    ua_string = UAString("Hello, World!")
    assert ua_string.xml_encode(include_xmlns=False) == "<String>Hello, World!</String>"
    assert (
        ua_string.xml_encode(include_xmlns=True)
        == '<String xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd">Hello, World!</String>'
    )

    assert UAString("The and symbol: & is escaped").xml_encode(include_xmlns=False) == (
        "<String>The and symbol: &amp; is escaped</String>"
    )


def test_ua_string_json_encode():
    # Test case 1: Value is None
    ua_string = UAString(None)
    assert ua_string.json_encode() is None

    # Test case 2: Value is not None
    ua_string = UAString("Hello, World!")
    expected_result = '"Hello, World!"'
    assert ua_string.json_encode() == expected_result

    # Test case 3: Value contains double quotes
    ua_string = UAString('This "quote" is escaped')
    expected_result = json.dumps('This "quote" is escaped')
    assert ua_string.json_encode() == expected_result

    # Test case 4: Value contains backslashes
    ua_string = UAString("C:\\Program Files\\")
    expected_result = json.dumps("C:\\Program Files\\")
    assert ua_string.json_encode() == expected_result

    # Test case 5: Value contains special characters
    ua_string = UAString("<html> &amp; </html>")
    expected_result = json.dumps("<html> &amp; </html>")
    assert ua_string.json_encode() == expected_result

    # Test case 6: Value contains Unicode characters
    ua_string = UAString("Résumé")
    expected_result = '"Résumé"'
    assert ua_string.json_encode() == expected_result


def test_ua_datetime_creation_with_incorrect_type():
    with pytest.raises(TypeError):
        UADateTime(value="foo")
    with pytest.raises(TypeError):
        UADateTime(value=None)


def test_ua_datetime_creation_with_value():
    # initializing datetime object with python datetime object
    datetime_obj = datetime(2021, 1, 1, 0, 0, 0)
    ua_datetime = UADateTime(value=datetime_obj)
    assert ua_datetime.value == datetime_obj


def test_ua_datetime_xml_encode_with_value():
    ua_datetime = UADateTime(value=datetime(2021, 1, 1, 0, 0, 0))
    expected_xml = "<DateTime>2021-01-01T00:00:00.000000Z</DateTime>"
    actual_xml = ua_datetime.xml_encode(include_xmlns=False)
    assert actual_xml == expected_xml


def test_ua_datetime_json_encode_with_value():
    # initializing datetime object with python datetime object
    ua_datetime = UADateTime(value=pd.Timestamp("2021-01-01 00:00:00"))
    expected_json = '"2021-01-01T00:00:00.000000Z"'
    actual_json = ua_datetime.json_encode()
    assert actual_json == expected_json


def test_ua_byte_string_creation_with_incorrect_type():
    with pytest.raises(TypeError):
        UAByteString(value=42)


def test_ua_byte_string_creation_with_none_value():
    ua_byte_string = UAByteString(None)
    assert isinstance(ua_byte_string.value, pd._libs.missing.NAType)


def test_ua_byte_string_xml_encode():
    ua_byte_string = UAByteString(b"Hello, world!")
    assert (
        ua_byte_string.xml_encode(include_xmlns=False)
        == "<ByteString>SGVsbG8sIHdvcmxkIQ==</ByteString>"
    )
    assert (
        ua_byte_string.xml_encode(include_xmlns=True)
        == '<ByteString xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd">SGVsbG8sIHdvcmxkIQ==</ByteString>'
    )


def test_ua_byte_string_json_encode():
    # Test case 1: Value is None
    obj = UAByteString(None)
    assert obj.json_encode() is None

    # Test case 2: Value is not None
    value = b"Hello, world!"
    obj = UAByteString(value)
    # expected_output = json.dumps(b64encode(value).decode("utf-8"), ensure_ascii=False)
    expected_output = '"SGVsbG8sIHdvcmxkIQ=="'
    actual_output = obj.json_encode()
    assert actual_output == expected_output


def test_ua_boolean_creation_with_incorrect_type():
    with pytest.raises(TypeError):
        UABoolean(value=42)
    with pytest.raises(TypeError):
        UABoolean(value="Some string")


def test_ua_boolean_creation_with_none_value():
    ua_boolean = UABoolean(None)
    assert ua_boolean.value is None


def test_ua_boolean_json_encode_with_value():
    ua_boolean = UABoolean(value=True)
    assert ua_boolean.json_encode() == "true"
    ua_boolean = UABoolean(value=False)
    assert ua_boolean.json_encode() == "false"


def test_ua_boolean_xml_encode_with_value():
    ua_boolean = UABoolean(value=True)
    assert ua_boolean.xml_encode(include_xmlns=False) == "<Boolean>true</Boolean>"
    ua_boolean = UABoolean(value=False)
    assert ua_boolean.xml_encode(include_xmlns=False) == "<Boolean>false</Boolean>"
    assert (
        ua_boolean.xml_encode(include_xmlns=True)
        == '<Boolean xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd">false</Boolean>'
    )


def test_ua_boolean_json_encode_with_none_value():
    ua_boolean = UABoolean(None)
    assert ua_boolean.json_encode() == None


def test_ua_xml_element_creation_with_incorrect_type():
    with pytest.raises(TypeError):
        UAXMLElement(value=42)
    with pytest.raises(TypeError):
        UAXMLElement(value=None)


def test_ua_xml_element_creation_with_value():
    ua_xml_element = UAXMLElement(value="<foo>bar</foo>")
    assert ua_xml_element.value == "<foo>bar</foo>"


def test_ua_xml_element_xml_encode_with_value():
    ua_xml_element = UAXMLElement(value="<foo>bar</foo>")
    expected_xml = "<foo>bar</foo>"
    actual_xml = ua_xml_element.xml_encode(include_xmlns=False)
    assert actual_xml == expected_xml

    # Value contains special characters
    ua_xml_element = UAXMLElement("<html> &amp; </html>")
    expected_result = "<html> &amp; </html>"
    assert ua_xml_element.xml_encode(include_xmlns=False) == expected_result


def test_ua_xml_element_json_encode_with_value():
    ua_xml_element = UAXMLElement(value="<foo>bar</foo>")
    expected_json = '"<foo>bar</foo>"'
    actual_json = ua_xml_element.json_encode()
    assert actual_json == expected_json

    # Value contains special characters
    ua_string = UAXMLElement("<html> &amp; </html>")
    expected_result = json.dumps("<html> &amp; </html>")
    assert ua_string.json_encode() == expected_result


def test_ua_node_id_creation_with_incorrect_type():
    with pytest.raises(ValueError):
        UANodeId(value=42, namespace=0, nodeid_type="foo")
    with pytest.raises(TypeError):
        UANodeId(value="foo", namespace=0, nodeid_type=0)
    with pytest.raises(TypeError):
        UANodeId(value="-21", namespace=0, nodeid_type=0)
    with pytest.raises(TypeError):
        UANodeId(value=42, namespace="foo", nodeid_type=1)
    with pytest.raises(ValueError):
        UANodeId(value="foo", namespace=1, nodeid_type="q")


def test_ua_node_id_creation_with_correct_values():
    ua_node_id = UANodeId(value="42", namespace=0, nodeid_type=0)
    assert ua_node_id.value == "42"
    assert ua_node_id.namespace == 0
    assert ua_node_id.nodeid_type == NodeIdType.NUMERIC
    assert isinstance(ua_node_id.nodeid_type, NodeIdType)

    ua_node_id = UANodeId(value="42", namespace=0, nodeid_type="i")
    assert ua_node_id.value == "42"
    assert ua_node_id.namespace == 0
    assert ua_node_id.nodeid_type == NodeIdType.NUMERIC
    assert isinstance(ua_node_id.nodeid_type, NodeIdType)

    ua_node_id = UANodeId(
        value="ParentBrowseName.MyBrowseName", namespace=0, nodeid_type=1
    )
    assert ua_node_id.value == "ParentBrowseName.MyBrowseName"
    assert ua_node_id.namespace == 0
    assert ua_node_id.nodeid_type == NodeIdType.STRING
    assert isinstance(ua_node_id.nodeid_type, NodeIdType)

    ua_node_id = UANodeId(
        value="ParentBrowseName.MyBrowseName", namespace=0, nodeid_type="s"
    )
    assert ua_node_id.value == "ParentBrowseName.MyBrowseName"
    assert ua_node_id.namespace == 0
    assert ua_node_id.nodeid_type == NodeIdType.STRING
    assert isinstance(ua_node_id.nodeid_type, NodeIdType)


def test_ua_node_id_xml_encode_with_value():
    ua_node_id = UANodeId(value="42", namespace=0, nodeid_type=0)
    expected_xml = "<Identifier>i=42</Identifier>"
    actual_xml = ua_node_id.xml_encode(include_xmlns=False)
    assert actual_xml == expected_xml

    ua_node_id = UANodeId(value="42", namespace=0, nodeid_type=0)
    expected_xml = '<Identifier xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd">i=42</Identifier>'
    actual_xml = ua_node_id.xml_encode(include_xmlns=True)
    assert actual_xml == expected_xml


def test_ua_node_id_json_encode_with_value():
    ua_node_id = UANodeId(value="42", namespace=0, nodeid_type=0)
    expected_json = '{"Id":42}'
    actual_json = ua_node_id.json_encode()
    assert actual_json == expected_json

    ua_node_id = UANodeId(value="42", namespace=1, nodeid_type=0)
    expected_json = '{"Namespace":1,"Id":42}'
    actual_json = ua_node_id.json_encode()
    assert actual_json == expected_json

    ua_node_id = UANodeId(value="MyNodeIdentifier", namespace=1, nodeid_type=1)
    expected_json = '{"Namespace":1,"IdType":1,"Id":"MyNodeIdentifier"}'
    actual_json = ua_node_id.json_encode()
    assert actual_json == expected_json


def test_ua_qualified_name_creation_with_incorrect_type():
    with pytest.raises(TypeError):
        UAQualifiedName(name=42, namespace_index=0)
    with pytest.raises(TypeError):
        UAQualifiedName(name="foo", namespace_index="bar")
    with pytest.raises(ValueError):
        UAQualifiedName(name="foo", namespace_index=-1)


def test_ua_qualified_name_creation_with_correct_values():
    ua_qualified_name = UAQualifiedName(name="foo", namespace_index=0)
    assert ua_qualified_name.name == "foo"
    assert ua_qualified_name.namespace_index == 0

    ua_qualified_name = UAQualifiedName(name="foo", namespace_index=1)
    assert ua_qualified_name.name == "foo"
    assert ua_qualified_name.namespace_index == 1


def test_ua_qualified_name_xml_encode_with_value():
    ua_qualified_name = UAQualifiedName(name="foo", namespace_index=0)
    expected_xml = "<QualifiedName><NamespaceIndex>0</NamespaceIndex><Name>foo</Name></QualifiedName>"
    actual_xml = ua_qualified_name.xml_encode(include_xmlns=False)
    assert actual_xml == expected_xml

    ua_qualified_name = UAQualifiedName(name="foo", namespace_index=1)
    expected_xml = '<QualifiedName xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd"><NamespaceIndex>1</NamespaceIndex><Name>foo</Name></QualifiedName>'
    actual_xml = ua_qualified_name.xml_encode(include_xmlns=True)
    assert actual_xml == expected_xml


def test_ua_qualified_name_json_encode_with_value():
    ua_qualified_name = UAQualifiedName(name="foo", namespace_index=0)
    expected_json = '{"Name":"foo"}'
    actual_json = ua_qualified_name.json_encode()
    assert actual_json == expected_json

    ua_qualified_name = UAQualifiedName(name="foo", namespace_index=1)
    expected_json = '{"Name":"foo","Uri":1}'
    actual_json = ua_qualified_name.json_encode()
    assert actual_json == expected_json


def test_ua_localized_text_creation_with_incorrect_type():
    with pytest.raises(TypeError):
        UALocalizedText(text=42, locale="en-US")
    with pytest.raises(TypeError):
        UALocalizedText(text="foo", locale=42)
    with pytest.raises(ValueError):
        UALocalizedText(text="foo", locale="foobarillo")


def test_ua_localized_text_creation_with_correct_values():
    ua_localized_text = UALocalizedText(text=None, locale=None)
    assert ua_localized_text.text is None
    assert ua_localized_text.locale is None

    ua_localized_text = UALocalizedText(text="foo", locale="en-US")
    assert ua_localized_text.text == "foo"
    assert ua_localized_text.locale == "en-US"

    ua_localized_text = UALocalizedText(text="foo", locale="de-DE")
    assert ua_localized_text.text == "foo"
    assert ua_localized_text.locale == "de-DE"


def test_ua_localized_text_is_valid_locale_with_different_locales():
    valid_locales = [
        "en-US",
        "de-DE",
        "fr-FR",
        "es-ES",
        "zh-CN",
        "zh-TW",
        "ja",
        "ko",
        "itn",
        "pt-BRA_kls",
        "pt-BRr-kls.ks",
        "pt-BRr-kls.ksslbei",
    ]
    for locale in valid_locales:
        assert UALocalizedText.is_valid_locale(locale) is True

    invalid_locales = [
        "foobarillo",
        "engl",
        "dech",
        "fr--FR",
        "es__ES",
        "zh-CNNR",
        "zh_CNNR",
    ]
    for locale in invalid_locales:
        assert UALocalizedText.is_valid_locale(locale) is False


def test_ua_localized_text_xml_encoding_with_value():
    ua_localized_text = UALocalizedText(text="foo", locale="en-US")
    expected_xml = (
        "<LocalizedText><Locale>en-US</Locale><Text>foo</Text></LocalizedText>"
    )
    actual_xml = ua_localized_text.xml_encode(include_xmlns=False)
    assert actual_xml == expected_xml

    ua_localized_text = UALocalizedText(text="foo", locale="en-US")
    expected_xml = '<LocalizedText xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd"><Locale>en-US</Locale><Text>foo</Text></LocalizedText>'
    actual_xml = ua_localized_text.xml_encode(include_xmlns=True)
    assert actual_xml == expected_xml


def test_ua_localized_text_json_encoding_with_none():
    ua_localized_text = UALocalizedText(text=None, locale=None)
    expected_json = '{"Text":""}'
    actual_json = ua_localized_text.json_encode()
    assert actual_json == expected_json

    ua_localized_text = UALocalizedText(text=None, locale="en-US")
    expected_json = '{"Text":"","Locale":"en-US"}'
    actual_json = ua_localized_text.json_encode()
    assert actual_json == expected_json


def test_ua_localized_text_json_encoding_with_pd_na():
    ua_localized_text = UALocalizedText(text=pd.NA, locale=pd.NA)
    expected_json = '{"Text":""}'
    actual_json = ua_localized_text.json_encode()
    assert actual_json == expected_json

    ua_localized_text = UALocalizedText(text=pd.NA, locale="en-US")
    expected_json = '{"Text":"","Locale":"en-US"}'
    actual_json = ua_localized_text.json_encode()
    assert actual_json == expected_json


def test_ua_localized_text_json_encoding_with_input_locale_as_pd_na():
    ua_localized_text = UALocalizedText(text="Some text", locale=pd.NA)
    expected_json = '{"Text":"Some text"}'
    actual_json = ua_localized_text.json_encode(input_locale=pd.NA)
    assert actual_json == expected_json


def test_ua_localized_text_json_encoding_with_locale_as_pd_na():
    ua_localized_text = UALocalizedText(text="Some text", locale=pd.NA)
    expected_json = '{"Text":"Some text"}'
    actual_json = ua_localized_text.json_encode()
    assert actual_json == expected_json


def test_ua_localized_text_json_encode_with_input_locale():
    # Testing that input_locale overrides the locale if there
    ua_localized_text = UALocalizedText(text="foo", locale="en-US")
    expected_json = '{"Text":"foo","Locale":"de-DE"}'
    actual_json = ua_localized_text.json_encode(input_locale="de-DE")
    assert actual_json == expected_json

    # Testing that input_locale overrides the locale if None
    ua_localized_text = UALocalizedText(text="foo", locale=None)
    expected_json = '{"Text":"foo","Locale":"de-DE"}'
    actual_json = ua_localized_text.json_encode(input_locale="de-DE")
    assert actual_json == expected_json


def test_ua_localized_text_json_encode_with_value():
    ua_localized_text = UALocalizedText(text="foo", locale="en-US")
    expected_json = '{"Text":"foo","Locale":"en-US"}'
    actual_json = ua_localized_text.json_encode()
    assert actual_json == expected_json

    ua_localized_text = UALocalizedText(text="foo", locale="de-DE")
    expected_json = '{"Text":"foo","Locale":"de-DE"}'
    actual_json = ua_localized_text.json_encode()
    assert actual_json == expected_json


def test_ua_extension_object_creation_with_incorrect_type():
    with pytest.raises(TypeError):
        UAExtensionObject(
            body=UAString("foo"),
            type_nodeid=UANodeId(namespace=0, nodeid_type=NodeIdType.NUMERIC, value=1),
        )
    with pytest.raises(TypeError):
        UAExtensionObject(
            body=UAStructure(UAString("foo")),
            type_nodeid=45,
        )


def test_ua_extension_object_creation_with_correct_values():
    ua_extension_object = UAExtensionObject(
        body=UAStructure(UAString("foo")),
        type_nodeid=UANodeId(namespace=0, nodeid_type=NodeIdType.NUMERIC, value=1),
    )
    assert ua_extension_object.body == UAStructure(UAString("foo"))
    assert ua_extension_object.type_nodeid == UANodeId(
        namespace=0, nodeid_type=NodeIdType.NUMERIC, value=1
    )

    ua_extension_object = UAExtensionObject(
        body=UAStructure(UABoolean(True)),
        type_nodeid=UANodeId(namespace=0, nodeid_type=NodeIdType.NUMERIC, value=1),
    )
    assert ua_extension_object.body == UAStructure(UABoolean(True))
    assert ua_extension_object.type_nodeid == UANodeId(
        namespace=0, nodeid_type=NodeIdType.NUMERIC, value=1
    )


def test_ua_extention_object_xml_encode():
    ua_extension_object = UAExtensionObject(
        body=UAStructure(UAString("foo")),
        type_nodeid=UANodeId(namespace=0, nodeid_type=NodeIdType.NUMERIC, value=1),
    )
    expected_xml = "<ExtensionObject><TypeId><Identifier>i=1</Identifier></TypeId><Body><String>foo</String></Body></ExtensionObject>"
    actual_xml = ua_extension_object.xml_encode(include_xmlns=False)
    assert actual_xml == expected_xml

    ua_extension_object = UAExtensionObject(
        body=UAStructure(UABoolean(True)),
        type_nodeid=UANodeId(namespace=0, nodeid_type=NodeIdType.NUMERIC, value=1),
    )
    actual_xml = ua_extension_object.xml_encode(include_xmlns=True)
    expected_xml = '<ExtensionObject xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd"><TypeId><Identifier>i=1</Identifier></TypeId><Body><Boolean>true</Boolean></Body></ExtensionObject>'
    assert actual_xml == expected_xml


def test_ua_extension_object_json_encode():
    ua_extension_object = UAExtensionObject(
        body=UAStructure(UAString("foo")),
        type_nodeid=UANodeId(namespace=0, nodeid_type=NodeIdType.NUMERIC, value=1),
    )
    expected_json = '{"TypeId":{"Id":1},"Body":"foo"}'
    actual_json = ua_extension_object.json_encode()
    assert actual_json == expected_json

    ua_extension_object = UAExtensionObject(
        body=UAStructure(UABoolean(True)),
        type_nodeid=UANodeId(namespace=0, nodeid_type=NodeIdType.NUMERIC, value=1),
    )
    expected_json = '{"TypeId":{"Id":1},"Body":true}'
    actual_json = ua_extension_object.json_encode()
    assert actual_json == expected_json


def test_ua_variant_creation_with_incorrect_type():
    with pytest.raises(TypeError):
        UAVariant(value="foo", type="int")
    with pytest.raises(TypeError):
        UAVariant(value=42, type=42)
    with pytest.raises(TypeError):
        UAVariant(value="foo", type="foobarillo")

    with pytest.raises(TypeError):
        UAVariant(value=None, type=None)
    with pytest.raises(TypeError):
        UAVariant(value=42, type=VariantType.Int16)
    with pytest.raises(TypeError):
        UAVariant(value="somestring", type=VariantType.String)


def test_ua_variant_creation_with_correct_values():
    ua_variant = UAVariant(value=None)
    assert ua_variant.value is None
    assert ua_variant.type is VariantType.Null

    ua_variant = UAVariant(value=UAUInt16(3), type=VariantType.UInt16)
    assert ua_variant.value == UAUInt16(3)
    assert ua_variant.type is VariantType.UInt16

    ua_variant = UAVariant(value=UAInt16(-3), type=VariantType.Int16)
    assert ua_variant.value == UAInt16(-3)
    assert ua_variant.type is VariantType.Int16

    ua_variant = UAVariant(value=UAInt32(-3), type=VariantType.Int32)
    assert ua_variant.value == UAInt32(-3)
    assert ua_variant.type is VariantType.Int32

    ua_variant = UAVariant(value=UAUInt32(3), type=VariantType.UInt32)
    assert ua_variant.value == UAUInt32(3)
    assert ua_variant.type is VariantType.UInt32

    ua_variant = UAVariant(value=UAUInt64(0), type=VariantType.UInt64)
    assert ua_variant.value == UAUInt64(0)
    assert ua_variant.type is VariantType.UInt64

    ua_variant = UAVariant(value=UAInt64(0), type=VariantType.Int64)
    assert ua_variant.value == UAInt64(0)
    assert ua_variant.type is VariantType.Int64

    ua_variant = UAVariant(value=UAString("foo"), type=VariantType.String)
    assert ua_variant.value == UAString("foo")
    assert ua_variant.type is VariantType.String

    ua_variant = UAVariant(value=UABoolean(True), type=VariantType.Boolean)
    assert ua_variant.value == UABoolean(True)
    assert ua_variant.type is VariantType.Boolean

    ua_variant = UAVariant(value=UAByteString(b"foo"), type=VariantType.ByteString)
    assert ua_variant.value == UAByteString(b"foo")
    assert ua_variant.type is VariantType.ByteString

    ua_variant = UAVariant(value=UAFloat(3.14), type=VariantType.Float)
    assert ua_variant.value == UAFloat(3.14)
    assert ua_variant.type is VariantType.Float

    ua_variant = UAVariant(value=UADouble(3.14), type=VariantType.Double)
    assert ua_variant.value == UADouble(3.14)
    assert ua_variant.type is VariantType.Double

    ua_variant = UAVariant(
        value=UADateTime(value=pd.Timestamp("2021-01-01 00:00:00")),
        type=VariantType.DateTime,
    )
    assert ua_variant.value == UADateTime(value=pd.Timestamp("2021-01-01 00:00:00"))
    assert ua_variant.type is VariantType.DateTime

    # ua_variant = UAVariant(value=UAGuid("3.14"), type=VariantType.Guid)
    # assert ua_variant.value == UAGuid("3.14")
    # assert ua_variant.type is VariantType.Guid

    ua_variant = UAVariant(
        value=UALocalizedText(text="foo", locale="en-US"),
        type=VariantType.LocalizedText,
    )
    assert ua_variant.value == UALocalizedText("foo", "en-US")
    assert ua_variant.type is VariantType.LocalizedText

    ua_variant = UAVariant(
        value=UAQualifiedName(name="foo", namespace_index=1),
        type=VariantType.QualifiedName,
    )
    assert ua_variant.value == UAQualifiedName(name="foo", namespace_index=1)
    assert ua_variant.type is VariantType.QualifiedName

    # ua_variant = UAVariant(value=UAStatusCode(0), type=VariantType.StatusCode)
    # assert ua_variant.value == UAStatusCode(0)
    # assert ua_variant.type is VariantType.StatusCode

    ua_variant = UAVariant(value=UAByte(0), type=VariantType.Byte)
    assert ua_variant.value == UAByte(0)
    assert ua_variant.type is VariantType.Byte

    ua_variant = UAVariant(value=UASByte(0), type=VariantType.SByte)
    assert ua_variant.value == UASByte(0)
    assert ua_variant.type is VariantType.SByte

    # ua_variant = UAVariant(value=UADataValue(0), type=VariantType.DataValue)
    # assert ua_variant.value == UADataValue(0)
    # assert ua_variant.type is VariantType.DataValue

    # ua_variant = UAVariant(value=UAExpandedNodeId(0), type=VariantType.ExpandedNodeId)
    # assert ua_variant.value == UAExpandedNodeId(0)
    # assert ua_variant.type is VariantType.ExpandedNodeId

    # ua_variant = UAVariant(value=UAStatusCode(0), type=VariantType.Variant)
    # assert ua_variant.value == UAStatusCode(0)
    # assert ua_variant.type is VariantType.Variant


def test_ua_variant_creation_with_correct_values_without_setting_type():
    ua_variant = UAVariant(value=UAUInt16(3))
    assert ua_variant.value == UAUInt16(3)
    assert ua_variant.type is VariantType.UInt16

    ua_variant = UAVariant(value=UAString("foo"))
    assert ua_variant.value == UAString("foo")
    assert ua_variant.type is VariantType.String

    ua_variant = UAVariant(value=UABoolean(True))
    assert ua_variant.value == UABoolean(True)
    assert ua_variant.type is VariantType.Boolean

    ua_variant = UAVariant(value=UAByteString(b"foo"))
    assert ua_variant.value == UAByteString(b"foo")
    assert ua_variant.type is VariantType.ByteString

    ua_variant = UAVariant(value=UAFloat(3.14))
    assert ua_variant.value == UAFloat(3.14)
    assert ua_variant.type is VariantType.Float

    ua_variant = UAVariant(value=UADouble(3.14))
    assert ua_variant.value == UADouble(3.14)
    assert ua_variant.type is VariantType.Double


def test_ua_variant_creation_with_different_range_objects():
    # Setting a UAExtensionObject for EURange
    ua_extension_object = UAExtensionObject(
        body=UAStructure(UAEURange(low=0, high=100)),
        type_nodeid=UANodeId(namespace=0, nodeid_type=NodeIdType.NUMERIC, value="885"),
    )
    ua_variant = UAVariant(ua_extension_object, type=VariantType.ExtensionObject)
    assert ua_variant.value == ua_extension_object
    assert ua_variant.type is VariantType.ExtensionObject

    # Setting a UAExtensionObject with UAStructure with UARange object
    ua_extension_object = UAExtensionObject(
        body=UAStructure(UARange(low=0, high=100)),
        type_nodeid=UANodeId(namespace=0, nodeid_type=NodeIdType.NUMERIC, value="885"),
    )
    ua_variant = UAVariant(ua_extension_object, type=VariantType.ExtensionObject)
    assert ua_variant.value == ua_extension_object
    assert ua_variant.type is VariantType.ExtensionObject

    # Setting a UAExtensionObject from EURange object
    ua_eu_range = UAEURange(low=0, high=100)
    ua_extension_object = UAExtensionObject(
        body=UAStructure(ua_eu_range),
        type_nodeid=UANodeId(namespace=0, nodeid_type=NodeIdType.NUMERIC, value="885"),
    )
    ua_variant = UAVariant(ua_extension_object, type=VariantType.ExtensionObject)
    assert ua_variant.value == ua_extension_object
    assert ua_variant.type is VariantType.ExtensionObject


def test_ua_variant_creation_with_different_engineering_unit_objects():
    ua_extension_object_value = UAExtensionObject(
        body=UAStructure(
            UAEUInformation(
                display_name=UALocalizedText("foo", "en-US"),
                description=UALocalizedText("bar", "en-US"),
                unit_id=42,
                namespace_uri="www.mynamespace.com/",
            ),
        ),
        type_nodeid=UANodeId(namespace=0, nodeid_type=NodeIdType.NUMERIC, value="887"),
    )
    ua_variant = UAVariant(
        value=ua_extension_object_value,
        type=VariantType.ExtensionObject,
    )
    assert ua_variant.value == ua_extension_object_value
    assert ua_variant.type is VariantType.ExtensionObject

    ua_engingeering_units = UAEngineeringUnits(
        display_name=UALocalizedText("foo", "en-US"),
        description=UALocalizedText("bar", "en-US"),
        unit_id=42,
        namespace_uri="www.mynamespace.com/",
    )
    ua_extension_object = UAExtensionObject(
        body=UAStructure(ua_engingeering_units),
        type_nodeid=UANodeId(namespace=0, nodeid_type=NodeIdType.NUMERIC, value="887"),
    )
    ua_variant = UAVariant(
        value=ua_extension_object,
        type=VariantType.ExtensionObject,
    )
    assert ua_variant.value == ua_variant.value
    assert ua_variant.type is VariantType.ExtensionObject


def test_ua_variant_xml_encode_with_value():
    ua_variant = UAVariant(value=UAString("MyStringaling"), type=VariantType.String)
    actual_xml = ua_variant.xml_encode(include_xmlns=False)
    expected_xml = "<Variant><Value><String>MyStringaling</String></Value></Variant>"
    assert actual_xml == expected_xml

    ua_variant = UAVariant(value=UAFloat(3.14), type=VariantType.Float)
    expected_xml = (
        '<Variant xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd">'
        '<Value><Float xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd">3.14</Float></Value>'
        "</Variant>"
    )
    actual_xml = ua_variant.xml_encode(include_xmlns=True)
    assert actual_xml == expected_xml


def test_ua_variant_json_encode_with_value():
    ua_variant = UAVariant(value=UAFloat(3.14), type=VariantType.Float)
    expected_json = '{"Type":10,"Body":3.14}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    ua_variant = UAVariant(value=UAString("foo"), type=VariantType.String)
    expected_json = '{"Type":12,"Body":"foo"}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    ua_variant = UAVariant(value=UABoolean(True), type=VariantType.Boolean)
    expected_json = '{"Type":1,"Body":true}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    ua_variant = UAVariant(value=UAByte(0x01), type=VariantType.Byte)
    expected_json = '{"Type":3,"Body":1}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    ua_variant = UAVariant(value=UASByte(0), type=VariantType.SByte)
    expected_json = '{"Type":2,"Body":0}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    ua_variant = UAVariant(value=UAInt16(153), type=VariantType.Int16)
    expected_json = '{"Type":4,"Body":153}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    ua_variant = UAVariant(value=UAInt32(153), type=VariantType.Int32)
    expected_json = '{"Type":6,"Body":153}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    ua_variant = UAVariant(value=UAInt64(153), type=VariantType.Int64)
    expected_json = '{"Type":8,"Body":"153.0"}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    ua_variant = UAVariant(value=UAUInt64(153), type=VariantType.UInt64)
    expected_json = '{"Type":9,"Body":"153.0"}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    ua_variant = UAVariant(value=UADouble(3.14), type=VariantType.Double)
    expected_json = '{"Type":11,"Body":3.14}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    ua_variant = UAVariant(
        value=UADateTime(value=pd.Timestamp("2021-01-01 00:00:00")),
        type=VariantType.DateTime,
    )
    expected_json = '{"Type":13,"Body":"2021-01-01T00:00:00.000000Z"}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    # ua_variant = UAVariant(
    #     value=UAGuid("12345678-9ABC-DEF0-1234-56789ABCDEF0"), type=VariantType.Guid
    # )
    # expected_json = '{"Type":14,"Body":"12345678-9ABC-DEF0-1234-56789ABCDEF0"}'
    # actual_json = ua_variant.json_encode()
    # assert actual_json == expected_json

    ua_variant = UAVariant(value=UAByteString(b"foo"), type=VariantType.ByteString)
    expected_json = '{"Type":15,"Body":"Zm9v"}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    ua_variant = UAVariant(
        value=UAXMLElement("<foo>bar</foo>"), type=VariantType.XmlElement
    )
    expected_json = '{"Type":16,"Body":"<foo>bar</foo>"}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    ua_variant = UAVariant(
        value=UANodeId(namespace=0, nodeid_type=NodeIdType.NUMERIC, value=42),
        type=VariantType.NodeId,
    )
    expected_json = '{"Type":17,"Body":{"Id":42}}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    # ua_variant = UAVariant(value=UAExpandedNodeId(namespace=0, nodeid_type=NodeIdType.NUMERIC, value="42"), type=VariantType.ExpandedNodeId)
    # expected_json = '{"Type":18,"Body":{"Namespace":0,"IdType":0,"Value":42}}'
    # actual_json = ua_variant.json_encode()
    # assert actual_json == expected_json

    # ua_variant = UAVariant(value=UAStatusCode(0), type=VariantType.StatusCode)
    # expected_json = '{"Type":19,"Body":0}'
    # actual_json = ua_variant.json_encode()
    # assert actual_json == expected_json

    ua_variant = UAVariant(
        value=UAQualifiedName(namespace_index=1, name="foo"),
        type=VariantType.QualifiedName,
    )
    expected_json = '{"Type":20,"Body":{"Name":"foo","Uri":1}}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    ua_variant = UAVariant(
        value=UALocalizedText(locale="en", text="foo"), type=VariantType.LocalizedText
    )
    expected_json = '{"Type":21,"Body":{"Text":"foo","Locale":"en"}}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    ua_variant = UAVariant(
        value=UAExtensionObject(
            body=UAStructure(UAString("foo")),
            type_nodeid=UANodeId(namespace=0, nodeid_type=NodeIdType.NUMERIC, value=12),
        ),
        type=VariantType.ExtensionObject,
    )
    expected_json = '{"Type":22,"Body":{"TypeId":{"Id":12},"Body":"foo"}}'
    actual_json = ua_variant.json_encode()
    assert actual_json == expected_json

    # ua_variant = UAVariant(value=UADataValue(), type=VariantType.DataValue)
    # expected_json = '{"Type":23,"Body":{"Value":{"Type":0,"Body":null},"StatusCode":0,"SourceTimestamp":"1970-01-01T00:00:00Z","ServerTimestamp":"1970-01-01T00:00:00Z","SourcePicoseconds":0,"ServerPicoseconds":0}}'
    # actual_json = ua_variant.json_encode()
    # assert actual_json == expected_json

    # ua_variant = UAVariant(value=UADiagnosticInfo(), type=VariantType.DiagnosticInfo)
    # expected_json = '{"Type":24,"Body":{"EncodingMask":0,"SymbolicId":0,"NamespaceUri":0,"LocalizedText":0,"Locale":0,"AdditionalInfo":0,"InnerStatusCode":0,"InnerDiagnosticInfo":0}}'
    # actual_json = ua_variant.json_encode()
    # assert actual_json == expected_json


def test_ua_eu_information_creation_with_incorrect_values():
    with pytest.raises(TypeError):
        UAEUInformation(
            namespace_uri=42,
            unit_id=42,
            display_name="bar",
            description="bar",
        )
    with pytest.raises(TypeError):
        UAEUInformation(
            namespace_uri="www.mynamespace.com/",
            unit_id="42",
            display_name=UALocalizedText(text="bar", locale="en"),
            description=UALocalizedText(text="bar", locale="en"),
        )
    with pytest.raises(TypeError):
        UAEUInformation(
            namespace_uri="www.mynamespace.com/",
            unit_id=42,
            display_name=42,
            description=UALocalizedText(text="bar", locale="en"),
        )
    with pytest.raises(TypeError):
        UAEUInformation(
            namespace_uri="www.mynamespace.com/",
            unit_id=42,
            display_name=UALocalizedText(text="bar", locale="en"),
            description=42,
        )
    with pytest.raises(TypeError):
        UAEUInformation(
            namespace_namespace_uri="www.mynamespace.com/",
            unit_id=42,
            display_name=UALocalizedText(text="bar", locale="en"),
            description=UALocalizedText(text="bar", locale="en"),
        )


def test_ua_eu_information_creation_with_correct_values():
    ua_eu_information = UAEUInformation(
        namespace_uri="www.mynamespace.com/",
        unit_id=42,
        display_name=UALocalizedText(text="bar", locale="en"),
        description=UALocalizedText(text="bar", locale="en"),
    )
    assert ua_eu_information.namespace_uri == "www.mynamespace.com/"
    assert ua_eu_information.unit_id == 42
    assert ua_eu_information.display_name == UALocalizedText(text="bar", locale="en")
    assert ua_eu_information.description == UALocalizedText(text="bar", locale="en")


def test_ua_eu_information_creation_with_empty_description():
    expected_namespace_uri = "www.mynamespace.com/"
    expected_unit_id = 42
    expected_display_name = UALocalizedText(text="bar", locale="en")
    ua_eu_information = UAEUInformation(
        namespace_uri=expected_namespace_uri,
        unit_id=expected_unit_id,
        display_name=expected_display_name,
        description=pd.NA,
    )
    assert ua_eu_information.namespace_uri == expected_namespace_uri
    assert ua_eu_information.unit_id == expected_unit_id
    assert ua_eu_information.display_name == expected_display_name
    assert isinstance(ua_eu_information.description, pd._libs.missing.NAType)


def test_ua_eu_information_xml_encode():
    ua_eu_information = UAEUInformation(
        namespace_uri="www.mynamespace.com/",
        unit_id=42,
        display_name=UALocalizedText(text="bar", locale="en"),
        description=UALocalizedText(text="bar", locale="en"),
    )
    actual_xml = ua_eu_information.xml_encode(include_xmlns=False)
    expected_xml = "<EUInformation><NamespaceUri>www.mynamespace.com/</NamespaceUri><UnitId>42</UnitId><DisplayName><Locale>en</Locale><Text>bar</Text></DisplayName><Description><Locale>en</Locale><Text>bar</Text></Description></EUInformation>"
    assert actual_xml == expected_xml


def test_ua_eu_information_xml_encode_with_xmlns():
    ua_eu_information = UAEUInformation(
        namespace_uri="www.mynamespace.com/",
        unit_id=42,
        display_name=UALocalizedText(text="bar", locale="en"),
        description=UALocalizedText(text="bar", locale="en"),
    )
    actual_xml = ua_eu_information.xml_encode(include_xmlns=True)
    expected_xml = '<EUInformation xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd"><NamespaceUri>www.mynamespace.com/</NamespaceUri><UnitId>42</UnitId><DisplayName><Locale>en</Locale><Text>bar</Text></DisplayName><Description><Locale>en</Locale><Text>bar</Text></Description></EUInformation>'
    assert actual_xml == expected_xml


def test_ua_eu_information_json_encode():
    ua_eu_information = UAEUInformation(
        namespace_uri="www.mynamespace.com/",
        unit_id=42,
        display_name=UALocalizedText(text="bar", locale="en"),
        description=UALocalizedText(text="bar", locale="en"),
    )
    actual_json = ua_eu_information.json_encode()
    expected_json = '{"DisplayName":{"Text":"bar","Locale":"en"},"Description":{"Text":"bar","Locale":"en"},"UnitId":42,"NamespaceUri":"www.mynamespace.com/"}'
    assert actual_json == expected_json


def test_ua_engineering_units_creation_with_incorrect_values():
    # Incorrect display name type
    with pytest.raises(TypeError):
        UAEngineeringUnits(
            display_name=42,
            description="bar",
            unit_id=42,
            namespace_uri="www.mynamespace.com/",
        )
    # Incorrect description type
    with pytest.raises(TypeError):
        UAEngineeringUnits(
            display_name="bar",
            description=42,
            unit_id=42,
            namespace_uri="www.mynamespace.com/",
        )
    # Incorrect unit id type
    with pytest.raises(TypeError):
        UAEngineeringUnits(
            display_name="bar",
            description="bar",
            unit_id="42",
            namespace_uri="www.mynamespace.com/",
        )
    # Incorrect namespace uri type
    with pytest.raises(TypeError):
        UAEngineeringUnits(
            display_name="bar",
            description="bar",
            unit_id=42,
            namespace_uri=42,
        )


def test_ua_engineering_units_creation_with_correct_values():
    ua_engineering_units = UAEngineeringUnits(
        display_name=UALocalizedText("bar", "en"),
        description=UALocalizedText("bar", "en-US"),
        unit_id=42,
        namespace_uri="www.mynamespace.com/",
    )
    assert ua_engineering_units.ua_eu_information.display_name == UALocalizedText(
        "bar", "en"
    )
    assert ua_engineering_units.ua_eu_information.description == UALocalizedText(
        "bar", "en-US"
    )
    assert ua_engineering_units.ua_eu_information.unit_id == 42
    assert (
        ua_engineering_units.ua_eu_information.namespace_uri == "www.mynamespace.com/"
    )


def test_ua_engineering_units_xml_encode_with_value():
    ua_engineering_units = UAEngineeringUnits(
        display_name=UALocalizedText("bar", "en"),
        description=UALocalizedText("bar", "en-US"),
        unit_id=42,
        namespace_uri="www.mynamespace.com/",
    )
    expected_xml = "<ExtensionObject><TypeId><Identifier>i=888</Identifier></TypeId><Body><EUInformation><NamespaceUri>www.mynamespace.com/</NamespaceUri><UnitId>42</UnitId><DisplayName><Locale>en</Locale><Text>bar</Text></DisplayName><Description><Locale>en-US</Locale><Text>bar</Text></Description></EUInformation></Body></ExtensionObject>"
    actual_xml = ua_engineering_units.xml_encode(include_xmlns=False)
    assert actual_xml == expected_xml

    expected_xml = '<ExtensionObject xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd"><TypeId><Identifier>i=888</Identifier></TypeId><Body><EUInformation><NamespaceUri>www.mynamespace.com/</NamespaceUri><UnitId>42</UnitId><DisplayName><Locale>en</Locale><Text>bar</Text></DisplayName><Description><Locale>en-US</Locale><Text>bar</Text></Description></EUInformation></Body></ExtensionObject>'
    actual_xml = ua_engineering_units.xml_encode(include_xmlns=True)
    assert actual_xml == expected_xml


def test_ua_engineering_units_json_encode_with_value():
    ua_engineering_units = UAEngineeringUnits(
        display_name=UALocalizedText("bar", "en"),
        description=UALocalizedText("bar", "en-US"),
        unit_id=42,
        namespace_uri="www.mynamespace.com/",
    )
    expected_json = '{"TypeId":{"Id":888},"Body":{"DisplayName":{"Text":"bar","Locale":"en"},"Description":{"Text":"bar","Locale":"en-US"},"UnitId":42,"NamespaceUri":"www.mynamespace.com/"}}'
    actual_json = ua_engineering_units.json_encode()
    assert actual_json == expected_json


def test_ua_range_creation_with_incorrect_values():
    with pytest.raises(TypeError):
        UARange(low="foo", high=100)
    with pytest.raises(TypeError):
        UARange(low=0, high="foo")
    with pytest.raises(ValueError):
        UARange(low=100, high=0)


def test_ua_range_creation_with_correct_values():
    ua_range = UARange(low=0, high=100)
    assert ua_range.low == 0
    assert ua_range.high == 100


def test_ua_range_xml_encode_with_value():
    ua_range = UARange(low=0, high=100)
    expected_xml = "<Range><Low>0.0</Low><High>100.0</High></Range>"
    actual_xml = ua_range.xml_encode(include_xmlns=False)
    assert actual_xml == expected_xml

    expected_xml = '<Range xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd"><Low>0.0</Low><High>100.0</High></Range>'
    actual_xml = ua_range.xml_encode(include_xmlns=True)
    assert actual_xml == expected_xml


def test_ua_range_json_encode_with_value():
    ua_range = UARange(low=0, high=100)
    expected_json = '{"Low":0.0,"High":100.0}'
    actual_json = ua_range.json_encode()
    assert actual_json == expected_json


def test_ua_eurange_creation_with_incorrect_values():
    with pytest.raises(TypeError):
        UAEURange(low="foo", high=100)
    with pytest.raises(TypeError):
        UAEURange(low=0, high="foo")
    with pytest.raises(ValueError):
        UAEURange(low=100, high=0)


def test_ua_eurange_creation_with_correct_values():
    ua_eurange = UAEURange(low=0, high=100)
    assert ua_eurange.ua_range.low == 0
    assert ua_eurange.ua_range.high == 100


def test_ua_eurange_xml_encode():
    ua_eurange = UAEURange(low=0, high=100)
    expected_xml = "<ExtensionObject><TypeId><Identifier>i=885</Identifier></TypeId><Body><Range><Low>0.0</Low><High>100.0</High></Range></Body></ExtensionObject>"
    actual_xml = ua_eurange.xml_encode(include_xmlns=False)
    assert actual_xml == expected_xml

    expected_xml = '<ExtensionObject xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd"><TypeId><Identifier>i=885</Identifier></TypeId><Body><Range><Low>0.0</Low><High>100.0</High></Range></Body></ExtensionObject>'
    actual_xml = ua_eurange.xml_encode(include_xmlns=True)
    assert actual_xml == expected_xml


def test_ua_eurange_json_encode_with_value():
    ua_eurange = UAEURange(low=0, high=100)
    expected_json = '{"TypeId":{"Id":885},"Body":{"Low":0.0,"High":100.0}}'
    actual_json = ua_eurange.json_encode()
    assert actual_json == expected_json



# Copyright 2021 Prediktor AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path

import pandas as pd
from definitions import get_project_root

import opcua_tools as ot
from opcua_tools.nodeset_parser import (
    exclude_files_not_in_namespaces,
    get_list_of_xml_files,
    get_xml_namespaces,
)

PATH_HERE = os.path.dirname(__file__)


def test_parsing_without_errors():
    ot.parse_xml_dir(PATH_HERE + "/testdata/parser")


def test_parse_nodeid():
    s = "ns=1;i=0"
    nid = ot.parse_nodeid(s)

    assert type(nid.value) == str


def test_parser_easy_example_stays_put():
    gr = ot.UAGraph.from_path(PATH_HERE + "/testdata/generator")
    nodes = gr.get_normalized_nodes_df()
    references = gr.get_normalized_references_df()

    nodes.to_csv(PATH_HERE + "/expected/parser/nodes.csv", index=False)
    references.to_csv(PATH_HERE + "/expected/parser/references.csv", index=False)

    actual_nodes = pd.read_csv(PATH_HERE + "/expected/parser/nodes.csv")
    actual_references = pd.read_csv(PATH_HERE + "/expected/parser/references.csv")

    expected_nodes = pd.read_csv(
        PATH_HERE + "/expected/parser/expected_nodes.csv",
        dtype={"IsAbstract": bool, "Symmetric": bool},
    )
    expected_references = pd.read_csv(
        PATH_HERE + "/expected/parser/expected_references.csv"
    )

    pd.testing.assert_frame_equal(expected_references, actual_references)

    for c in expected_nodes.columns:
        if c != "Value":
            pd.testing.assert_series_equal(expected_nodes[c], actual_nodes[c])
        else:
            # Todo fix test error for value between linux/windows
            pass


def test_parse_file_list():
    path_to_xmls = str(Path(PATH_HERE) / "testdata" / "parser")
    xml_list = []
    for xml in os.listdir(path_to_xmls):
        full_xml_path = os.path.join(path_to_xmls, xml)
        if os.path.isfile(full_xml_path):
            xml_list.append(full_xml_path)

    parse_dict = ot.parse_xml_files(xml_list)
    assert "nodes" in parse_dict
    assert "references" in parse_dict
    assert "namespaces" in parse_dict
    assert "lookup_df" in parse_dict
    assert "models" in parse_dict


def test_parse_no_namespace_list(paper_example_path):
    path_to_xmls = str(paper_example_path)
    ua_graph = ot.UAGraph.from_path(path_to_xmls)

    # Checking that content is in correct namespace
    i_scd_ca_row = ua_graph.nodes[ua_graph.nodes["DisplayName"] == "I_SCD_CA"]
    oil_and_gas_system_row = ua_graph.nodes[
        ua_graph.nodes["DisplayName"] == "OilAndGasSystemType"
    ]
    site1_row = ua_graph.nodes[ua_graph.nodes["DisplayName"] == "Site1"]

    assert i_scd_ca_row["ns"].values[0] == 3
    assert oil_and_gas_system_row["ns"].values[0] == 2
    assert site1_row["ns"].values[0] == 1


def test_parse_regular_namespace_list(paper_example_path):
    namespace_dict = {
        0: "http://opcfoundation.org/UA/",
        1: "http://prediktor.no/apis/ua/",
        2: "http://prediktor.com/paper_example",
        3: "http://prediktor.com/iec63131_fragment",
        4: "http://prediktor.com/RDS-OG-Fragment",
    }
    path_to_xmls = str(paper_example_path)
    ua_graph = ot.UAGraph.from_path(path_to_xmls, namespace_dict)

    # Checking that content is in correct namespace
    i_scd_ca_row = ua_graph.nodes[ua_graph.nodes["DisplayName"] == "I_SCD_CA"]
    oil_and_gas_system_row = ua_graph.nodes[
        ua_graph.nodes["DisplayName"] == "OilAndGasSystemType"
    ]
    site1_row = ua_graph.nodes[ua_graph.nodes["DisplayName"] == "Site1"]

    assert i_scd_ca_row["ns"].values[0] == 3
    assert oil_and_gas_system_row["ns"].values[0] == 4
    assert site1_row["ns"].values[0] == 2


def test_parse_scattered_namespace_list(paper_example_path):
    namespace_dict = {
        0: "http://opcfoundation.org/UA/",
        7: "http://prediktor.no/apis/ua/",
        8: "http://prediktor.com/paper_example",
        3: "http://prediktor.com/iec63131_fragment",
        6: "http://prediktor.com/RDS-OG-Fragment",
    }
    path_to_xmls = str(paper_example_path)
    ua_graph = ot.UAGraph.from_path(path_to_xmls, namespace_dict)

    # Checking that content is in correct namespace
    i_scd_ca_row = ua_graph.nodes[ua_graph.nodes["DisplayName"] == "I_SCD_CA"]
    oil_and_gas_system_row = ua_graph.nodes[
        ua_graph.nodes["DisplayName"] == "OilAndGasSystemType"
    ]
    site1_row = ua_graph.nodes[ua_graph.nodes["DisplayName"] == "Site1"]

    assert i_scd_ca_row["ns"].values[0] == 3
    assert oil_and_gas_system_row["ns"].values[0] == 6
    assert site1_row["ns"].values[0] == 8


def test_get_xml_namespaces():
    test_parser_data = get_project_root() / "tests" / "testdata" / "parser"

    files = [
        str(test_parser_data / "Opc.ISA95.NodeSet2.xml"),
        str(test_parser_data / "Opc.Ua.IEC61850-6.NodeSet2.xml"),
        str(test_parser_data / "Opc.Ua.IEC61850-7-4.NodeSet2.xml"),
    ]

    expected_namespaces = {
        "http://www.OPCFoundation.org/UA/2013/01/ISA95",
        "http://opcfoundation.org/UA/IEC61850-6",
        "http://opcfoundation.org/UA/IEC61850-7-4",
    }

    namespace_set = set()
    for xml_file in files:
        namespace_uri_list = get_xml_namespaces(xml_file)
        namespace_set = namespace_set.union(set(namespace_uri_list))

    assert expected_namespaces == namespace_set


def test_get_list_of_xml_files():
    xml_directory = str(get_project_root() / "tests" / "testdata" / "parser")
    expected_xml_files = {
        os.path.join(xml_directory, "Opc.ISA95.NodeSet2.xml"),
        os.path.join(xml_directory, "Opc.Ua.IEC61850-6.NodeSet2.xml"),
        os.path.join(xml_directory, "Opc.Ua.IEC61850-7-3.NodeSet2.xml"),
        os.path.join(xml_directory, "Opc.Ua.IEC61850-7-4.NodeSet2.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part10.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part11.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part12.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part13.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part14.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part17.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part19.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part3.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part4.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part5.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part8.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part9.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Services.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.xml"),
    }

    actual_xml_files = set(get_list_of_xml_files(xml_directory))
    d = expected_xml_files.difference(actual_xml_files)
    print(d)
    assert expected_xml_files == actual_xml_files


def test_namespace_dict_parse_xml_dict():
    xml_directory = str(get_project_root() / "tests" / "testdata" / "parser")
    input_namespaces = [
        "http://opcfoundation.org/UA/",
        "http://www.OPCFoundation.org/UA/2013/01/ISA95",
        "http://opcfoundation.org/UA/IEC61850-7-4",
    ]
    expected_files = {
        os.path.join(xml_directory, "Opc.ISA95.NodeSet2.xml"),
        os.path.join(xml_directory, "Opc.Ua.IEC61850-7-4.NodeSet2.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part10.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part11.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part12.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part13.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part14.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part17.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part19.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part3.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part4.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part5.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part8.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Part9.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.Services.xml"),
        os.path.join(xml_directory, "Opc.Ua.NodeSet2.xml"),
    }

    xml_files = get_list_of_xml_files(xml_directory)
    xml_files = exclude_files_not_in_namespaces(xml_files, input_namespaces)

    assert expected_files == set(xml_files)



# Copyright 2021 Prediktor AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opcua_tools.navigation import *
from opcua_tools.nodes_manipulation import transform_ints_to_enums
from opcua_tools.nodeset_generator import create_nodeset2_file, validate_nodeset2_file
from opcua_tools.nodeset_parser import (
    normalize_wrt_nodeid,
    parse_xml,
    parse_xml_dir,
    parse_xml_files,
)
from opcua_tools.ua_data_types import *
from opcua_tools.ua_graph import UAGraph



# Copyright 2021 Prediktor AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest
from definitions import get_project_root

import opcua_tools as ot

PATH_HERE = os.path.dirname(__file__)


@pytest.fixture
def circular_uag():
    namespace_dict = {
        0: "http://opcfoundation.org/UA/",
        1: "http://test.org/Circular/",
    }
    xml_directory = str(get_project_root() / "tests" / "testdata" / "circular")
    return ot.UAGraph.from_path(xml_directory, namespace_dict)


def test_get_normalized_references(circular_uag):
    uag = circular_uag
    normalized_refs = uag.get_normalized_references_df("http://test.org/Circular/")
    length = len(normalized_refs)
    assert length == 12


def test_find_circular_reference_nodes(circular_uag):
    uag = circular_uag
    circular_refs = uag.find_circular_reference_nodes("http://test.org/Circular/")
    assert len(circular_refs) == 4
    assert str(circular_refs["NodeId"].values[0]) == "ns=1;i=5006"
    assert str(circular_refs["NodeId"].values[1]) == "ns=1;i=5003"
    assert str(circular_refs["NodeId"].values[2]) == "ns=1;i=5005"
    assert str(circular_refs["NodeId"].values[3]) == "ns=1;i=5004"



import re

from opcua_tools import nodeset_generator, ua_models


def test_required_models_are_parsed_to_xml():
    model = ua_models.UAModel(
        model_uri="http://opcfoundation.org/UA/IEC61850-6",
        publication_date="2018-02-05T00:00:00Z",
        version="2.0",
        required_models=[
            ua_models.UARequiredModel(
                model_uri="http://opcfoundation.org/UA/IEC61850-7-3",
                publication_date="2023-01-01T00:00:00Z",
                version="2.0",
            ),
            ua_models.UARequiredModel(
                model_uri="http://opcfoundation.org/UA/",
                publication_date="2019-05-01T00:00:00Z",
                version="1.04",
            ),
        ],
    )

    required_models_xml = nodeset_generator.create_required_models(model)

    assert (
        required_models_xml
        == """
        <RequiredModel ModelUri="{}" Version="{}" PublicationDate="{}" />
        <RequiredModel ModelUri="{}" Version="{}" PublicationDate="{}" />
    """.format(
            model.required_models[0].model_uri,
            model.required_models[0].version,
            model.required_models[0].publication_date,
            model.required_models[1].model_uri,
            model.required_models[1].version,
            model.required_models[1].publication_date,
        )
    )


def test_required_models_are_parsed_when_no_publication_date_given():
    model = ua_models.UAModel(
        model_uri="http://opcfoundation.org/UA/IEC61850-6",
        publication_date="2018-02-05T00:00:00Z",
        version="2.0",
        required_models=[
            ua_models.UARequiredModel(
                model_uri="http://opcfoundation.org/UA/IEC61850-7-3",
                publication_date=None,
                version="2.0",
            ),
        ],
    )

    required_models_xml = nodeset_generator.create_required_models(model)
    xml_tag_date_pattern = re.compile(r'PublicationDate="\d+-\d+-\d+')
    match_object = xml_tag_date_pattern.search(required_models_xml)

    assert match_object is not None



# Copyright 2021 Prediktor AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import copy
import re
from typing import Dict, Optional

import lxml.etree as ET
import pandas as pd
from dateutil import parser

from opcua_tools.ua_data_types import (
    NodeIdType,
    UABoolean,
    UAByte,
    UAByteString,
    UADateTime,
    UADouble,
    UAEngineeringUnits,
    UAEURange,
    UAExtensionObject,
    UAFloat,
    UAGuid,
    UAInt16,
    UAInt32,
    UAInt64,
    UAListOf,
    UALocalizedText,
    UANodeId,
    UASByte,
    UAString,
    UAUInt16,
    UAUInt32,
    UAUInt64,
    UAXMLElement,
)

simple = {
    "Boolean",
    "SByte",
    "Byte",
    "Int16",
    "UInt16",
    "Int32",
    "UInt32",
    "Int64",
    "UInt64",
    "Float",
    "Double",
    "String",
    "DateTime",
    "Guid",
    "ByteString",
    "NodeId",
}

tagsplit = re.compile(r"({.*\})(.*)")
uaxsd = "{http://opcfoundation.org/UA/2008/02/Types.xsd}"


def parse_value(val):
    v = next(v for v in val)
    return parse_value_element(v)


def parse_value_element(v):
    tagtype = tagsplit.match(v.tag).group(2)

    if tagtype.startswith("ListOf"):
        return parse_list_value(v, tagtype)

    return parse_singular_value(v, tagtype)


def parse_list_value(val, tagtype):
    typename = re.fullmatch(r"^ListOf(.*)", tagtype).group(1)
    l = []
    for v in val:
        l.append(parse_value_element(v))
    return UAListOf(value=tuple(l), typename=typename)


def parse_singular_value(val, tagtype):
    if tagtype in simple:
        if val.text is not None:
            stripped = val.text.strip()
        else:
            stripped = None
        if tagtype == "SByte":
            if stripped:
                return UASByte(value=int(stripped))

            return UASByte(value=None)

        elif tagtype == "Byte":
            if stripped:
                return UAByte(value=int(stripped))

            return UAByte(value=None)

        elif tagtype == "Int16":
            if stripped:
                return UAInt16(value=int(stripped))

            return UAInt16(value=None)

        elif tagtype == "UInt16":
            if stripped:
                return UAUInt16(value=int(stripped))

            return UAUInt16(value=None)

        elif tagtype == "Int32":
            if stripped:
                return UAInt32(value=int(stripped))

            return UAInt32(value=None)

        elif tagtype == "UInt32":
            if stripped:
                return UAUInt32(value=int(stripped))

            return UAUInt32(value=None)

        elif tagtype == "Int64":
            if stripped:
                return UAInt64(value=int(stripped))

            return UAInt64(value=None)

        elif tagtype == "UInt64":
            if stripped:
                return UAUInt64(value=int(stripped))

            return UAUInt64(value=None)

        elif tagtype == "Float":
            if stripped:
                return UAFloat(value=float(stripped))

            return UAFloat(value=None)

        elif tagtype == "Double":
            if stripped:
                return UADouble(value=float(stripped))

            return UADouble(value=None)

        elif tagtype == "String":
            if stripped is not None:
                return UAString(value=stripped)

            return UAString(value=None)

        elif tagtype == "DateTime":
            parsed = parser.parse(stripped)
            return UADateTime(value=parsed)

        elif tagtype == "Boolean":
            if stripped == "":
                return UABoolean(value=None)

            return parse_boolean(stripped)

        elif tagtype == "ByteString":
            if stripped is not None:
                return UAByteString(value=base64.b64decode(stripped))

            return UAByteString(value=None)

        elif tagtype == "Guid":
            if stripped is not None:
                return UAGuid(value=stripped)

            return UAGuid(value=None)

        elif tagtype == "NodeId":
            if val.text is not None:
                return parse_nodeid(val.text)

            return None

    elif tagtype == "TypeId":
        ident = val.find(uaxsd + "Identifier")
        return parse_nodeid(ident.text)

    elif tagtype == "LocalizedText":
        return parse_localized_text(val)

    elif tagtype == "ExtensionObject":
        typeid = val.find(uaxsd + "TypeId")
        if typeid is None:
            parsed_typeid = None
        else:
            parsed_typeid = parse_value_element(typeid)

        body = val.find(uaxsd + "Body")

        if parsed_typeid is not None:
            if (
                parsed_typeid.namespace == 0
                and parsed_typeid.nodeid_type == NodeIdType.NUMERIC
                and parsed_typeid.value == "888"
            ):
                return parse_engineering_units(body)
            if (
                parsed_typeid.namespace == 0
                and parsed_typeid.nodeid_type == NodeIdType.NUMERIC
                and parsed_typeid.value == "885"
            ):
                return parse_eu_range(body)

        if body is not None and len(body) > 0:
            nextbody = next(n for n in body)
            if nextbody is None:
                parsednextbody = None
            else:
                parsednextbody = parse_value_element(nextbody)
        else:
            parsednextbody = None

        if parsednextbody is not None and parsed_typeid is not None:
            return UAExtensionObject(type_nodeid=parsed_typeid, body=parsednextbody)
        else:
            return None

    else:
        # Need to copy element, cleanup_namespaces only works if a copy is made
        val_copy = copy.copy(val)
        # Remove unused xml namespaces
        ET.cleanup_namespaces(val_copy)

        return UAXMLElement(value=ET.tostring(val_copy).decode("utf-8"))

    raise NotImplementedError(tagtype)


def parse_nodeid(
    nodeidstr: str,
    namespace_map: Optional[Dict[int, int]] = None,
    alias_map: Optional[Dict[str, UANodeId]] = None,
):
    if alias_map is not None and nodeidstr in alias_map:
        return alias_map[nodeidstr]

    withns = r"ns=(.*?);([isgb])=(.*)"
    fm_withns = re.fullmatch(withns, nodeidstr)
    if fm_withns is not None:
        ns = int(fm_withns.group(1))
        if namespace_map is not None:
            ns = namespace_map[ns]
        nodeid_type = NodeIdType(fm_withns.group(2))
        value = fm_withns.group(3)
    else:
        withoutns = r"([isgb])=(.*)"
        fm_withoutns = re.fullmatch(withoutns, nodeidstr)
        nodeid_type = NodeIdType(fm_withoutns.group(1))
        ns = 0
        value = fm_withoutns.group(2)

    return UANodeId(ns, nodeid_type, str(value))


def parse_localized_text(el):
    txt = el.find(uaxsd + "Text")
    if txt is None:
        text = pd.NA
    else:
        text = txt.text

    loc = el.find(uaxsd + "Locale")
    if loc is None:
        locale = pd.NA
    elif loc.text is None:
        locale = pd.NA
    elif type(loc.text) == str:
        loc_text_str = loc.text.strip()
        if loc_text_str == "":
            locale = pd.NA
        else:
            locale = loc.text
    else:
        raise ValueError("Unexpected type for localized text locale")

    return UALocalizedText(text=text, locale=locale)


def parse_engineering_units(el):
    if el is not None:
        euinfo = el.find(uaxsd + "EUInformation")
        if euinfo is not None:
            namespace_uri_tag = euinfo.find(uaxsd + "NamespaceUri")
            if namespace_uri_tag.text is not None:
                namespace_uri = namespace_uri_tag.text.rstrip()
                unit_id_tag = euinfo.find(uaxsd + "UnitId")
                unit_id = int(unit_id_tag.text.strip())

                display_name_tag = euinfo.find(uaxsd + "DisplayName")
                display_name = parse_localized_text(display_name_tag)
                description_tag = euinfo.find(uaxsd + "Description")
                description = parse_localized_text(description_tag)
                return UAEngineeringUnits(
                    display_name=display_name,
                    description=description,
                    unit_id=unit_id,
                    namespace_uri=namespace_uri,
                )


def parse_eu_range(el):
    if el is not None:
        range_tag = el.find(uaxsd + "Range")
        if range_tag is not None:
            low_tag = range_tag.find(uaxsd + "Low")
            low = float(low_tag.text.strip())
            high_tag = range_tag.find(uaxsd + "High")
            high = float(high_tag.text.strip())
            return UAEURange(
                low=low,
                high=high,
            )


def parse_boolean(string):
    if string is not None:
        if string in ["true", "True"]:
            return UABoolean(value=True)
        else:
            return UABoolean(value=False)



# Copyright 2021 Prediktor AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib

from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="opcua-tools",
    version="1.6.1",
    description="OPCUA Tools for Python using Pandas DataFrames",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/PrediktorAS/opcua-tools",
    author="Magnus Bakken",
    author_email="olav.pedersen@prediktor.com",
    license="Apache License 2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    packages=["opcua_tools", "opcua_tools.validator"],
    include_package_data=True,
    install_requires=[
        "lxml>=4.9.3",
        "lxml<5.0",
        "xmltodict>=0.13.0",
        "pandas>=2.0.0",
        "pandas<2.2.0",
        "scipy>=1.10.0",
        "scipy<2.0",
    ],
)



# Copyright 2022 Prediktor AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pandas as pd

import opcua_tools as ot
from opcua_tools.ua_data_types import NodeIdType, UANodeId

PATH_HERE = os.path.dirname(__file__)


def test_create_node_paths_by_reference_types(paper_example_path, ua_graph_data_path):
    ua_graph = ot.UAGraph.from_path(str(paper_example_path))

    hierarchical_references = [
        "HasComponent",
        "HasProperty",
        "HasChild",
        "Organizes",
        "ProductAspect",
    ]
    new_hierarchical_paths = ua_graph.create_node_paths_by_reference_types(
        root_node_browsename="Site1",
        hierarchical_reference_types=hierarchical_references,
    )
    # Writing node_paths to file in order to avoid differences in reading expected nodes
    output_nodes_paths_data_path = (
        ua_graph_data_path / "output" / "output_node_paths.csv"
    )
    new_hierarchical_paths.to_csv(
        str(output_nodes_paths_data_path),
        float_format="{:f}".format,
        encoding="utf-8",
    )

    expected_nodes_paths_data_path = (
        ua_graph_data_path / "expected" / "expected_node_paths.csv"
    )
    expected_nodes_paths = pd.read_csv(str(expected_nodes_paths_data_path))
    output_nodes_paths = pd.read_csv(str(output_nodes_paths_data_path))

    pd.testing.assert_frame_equal(expected_nodes_paths, output_nodes_paths)


def test_ua_graph_reference_type_by_browsename(paper_example_path):
    ua_graph = ot.UAGraph.from_path(str(paper_example_path))

    reference_types = ua_graph.reference_type_by_browsename("HasComponent")
    assert reference_types == 48


def test_ua_graph_object_type_by_browsename(paper_example_path):
    ua_graph = ot.UAGraph.from_path(str(paper_example_path))

    object_types = ua_graph.object_type_by_browsename("DataTypeSystemType")
    assert object_types == 69


def test_ua_graph_variable_type_by_browsename(paper_example_path):
    ua_graph = ot.UAGraph.from_path(str(paper_example_path))

    variable_types = ua_graph.variable_type_by_browsename("BaseVariableType")
    assert variable_types == 59


def test_ua_graph_data_type_by_browsename(paper_example_path):
    ua_graph = ot.UAGraph.from_path(str(paper_example_path))

    data_types = ua_graph.data_type_by_browsename("Number")
    assert data_types == 3


def test_ua_graph_object_by_browsename(paper_example_path):
    ua_graph = ot.UAGraph.from_path(str(paper_example_path))

    objects = ua_graph.object_by_browsename("Site1")
    assert objects == 3909


def test_ua_graph_nodeid_by_browsename_without_nodeclass(paper_example_path):
    ua_graph = ot.UAGraph.from_path(str(paper_example_path))

    nodeid = ua_graph.nodeid_by_browsename("I_SCD_CA")
    expected_nodeid = UANodeId(namespace=3, nodeid_type=NodeIdType("i"), value="1035")
    assert nodeid == expected_nodeid


def test_ua_graph_get_instances_with_type_info(paper_example_path):
    ua_graph = ot.UAGraph.from_path(str(paper_example_path))

    references = ua_graph.get_instances_with_type_info()
    # references.to_csv(
    #     os.path.join(PATH_HERE, "expected", "ua_graph", "references.csv"),
    #     index=False
    # )
    output_references_path = os.path.join(PATH_HERE, "output", "output_references.csv")
    references.to_csv(output_references_path, index=False)
    output_references = pd.read_csv(output_references_path)

    expected_references = pd.read_csv(
        os.path.join(PATH_HERE, "expected", "ua_graph", "references.csv")
    )

    pd.testing.assert_frame_equal(expected_references, output_references)



import gc
import logging
from datetime import datetime
from io import StringIO
from typing import Dict, List, Optional, Union

import pandas as pd

import opcua_tools.nodes_manipulation as nodes_manipulation
from opcua_tools import memory_optimizer, ua_models
from opcua_tools.navigation import (
    fast_transitive_closure,
    find_relatives,
    hierarchical_references,
    resolve_ids_from_browsenames,
)
from opcua_tools.nodeset_generator import (
    create_lookup_df,
    create_nodeset2_file,
    denormalize_nodes_nodeids,
    denormalize_references_nodeids,
)
from opcua_tools.nodeset_parser import parse_xml_dir, parse_xml_files
from opcua_tools.ua_data_types import UANodeId

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
memory_optimizer.replace_default_pandas_del_method()


class UAGraph:
    def __init__(
        self,
        nodes: pd.DataFrame,
        references: pd.DataFrame,
        namespaces: List[str],
        models: List[ua_models.UAModel],
    ):
        self.nodes = nodes
        self.nodes["id"] = self.nodes["id"].astype(pd.Int32Dtype())
        self.references = references
        self.references["Src"] = self.references["Src"].astype(pd.Int32Dtype())
        self.references["Trg"] = self.references["Trg"].astype(pd.Int32Dtype())
        self.references["ReferenceType"] = self.references["ReferenceType"].astype(
            pd.Int32Dtype()
        )
        self.namespaces = namespaces
        self.models = models

    @classmethod
    def from_path(
        cls, path: str, namespace_dict: Optional[Dict[int, str]] = None
    ) -> "UAGraph":
        if namespace_dict:
            namespace_list = UAGraph._get_namespace_list(namespace_dict)
            parse_dict = parse_xml_dir(path, namespace_list)
        else:
            parse_dict = parse_xml_dir(path)

        ua_graph = cls(
            nodes=parse_dict["nodes"],
            references=parse_dict["references"],
            namespaces=parse_dict["namespaces"],
            models=parse_dict["models"],
        )

        del parse_dict["nodes"]
        del parse_dict["references"]
        gc.collect()

        nodes_manipulation.transform_ints_to_enums(ua_graph)

        return ua_graph

    @classmethod
    def from_file_list(
        cls, file_list: List[str], namespace_dict: Optional[Dict[int, str]] = None
    ) -> "UAGraph":
        if namespace_dict:
            namespace_list = UAGraph._get_namespace_list(namespace_dict)
            parse_dict = parse_xml_files(file_list, namespace_list)
        else:
            parse_dict = parse_xml_files(file_list)

        return cls(
            nodes=parse_dict["nodes"],
            references=parse_dict["references"],
            namespaces=parse_dict["namespaces"],
            models=parse_dict["models"],
        )

    @staticmethod
    def _get_namespace_list(namespace_dict: dict) -> list:
        namespace_list = []
        max_namespace = max(namespace_dict.keys()) + 1
        for namespace_index in range(0, max_namespace):
            if namespace_index in namespace_dict.keys():
                namespace_list.append(namespace_dict[namespace_index])
            else:
                namespace_list.append("None")
        return namespace_list

    def all_references_of_type(self, browsename: str):
        return self.references[
            self.references["ReferenceType"]
            == self.reference_type_by_browsename(browsename)
        ].copy()

    def __ua_nodeclass_by_browsename(
        self, browsename: str, nodeclass: Optional[str] = None
    ) -> int:
        """Retrieves the internal id from the OPC UA graph which has a NodeClass `nodeclass`
        (UAVariableType in the UAGraph), for the given `browsename` input. If multiple nodes
        are found, an error is raised.

        Args:
            browsename (str): BrowseName of the node to find.
            nodeclass (Optional[str], optional): NodeClass of the node to find. Defaults to None.

        Raises:
            ValueError: If browsename is None or empty string.
            ValueError: If no node is found.
            ValueError: If multiple nodes are found.

        Returns:
            int: The internal id of the node.
        """
        if browsename is None or browsename == "":
            raise ValueError(
                f'"browsename" must not be None or empty string, should be BrowseName of {nodeclass}'
            )

        if nodeclass:
            nodeclass_nodes = self.nodes[self.nodes["NodeClass"] == f"UA{nodeclass}"]
        else:
            nodeclass_nodes = self.nodes

        nodeclass_ids = resolve_ids_from_browsenames(
            nodes=nodeclass_nodes, browsenames=[browsename]
        )
        if len(nodeclass_ids) == 0:
            raise ValueError(f"Could not find {nodeclass} " + browsename)
        elif len(nodeclass_ids) > 1:
            raise ValueError(f"Multiple hits for {nodeclass} " + browsename + " found.")
        else:
            nodeclass_id = nodeclass_ids.iloc[0]
        return int(nodeclass_id)

    def reference_type_by_browsename(self, browsename: str) -> int:
        """Retrieves the internal id from the OPC UA graph which has a NodeClass "ReferenceType"
        (UAReferenceType in the UAGraph), for the given `browsename` input. If multiple nodes
        are found, an error is raised.

        Args:
            browsename (str): BrowseName of the node to find.

        Returns:
            int: The internal id of the node.
        """
        return self.__ua_nodeclass_by_browsename(browsename, "ReferenceType")

    def object_type_by_browsename(self, browsename: str) -> int:
        """Retrieves the internal id from the OPC UA graph which has a NodeClass "ObjectType"
        (UAObjectType in the UAGraph), for the given `browsename` input. If multiple nodes
        are found, an error is raised.

        Args:
            browsename (str): BrowseName of the node to find.

        Returns:
            int: The internal id of the node.
        """
        return self.__ua_nodeclass_by_browsename(browsename, "ObjectType")

    def variable_type_by_browsename(self, browsename: str) -> int:
        """Retrieves the internal id from the OPC UA graph which has a NodeClass "VariableType"
        (UAVariableType in the UAGraph), for the given `browsename` input. If multiple nodes
        are found, an error is raised.

        Args:
            browsename (str): BrowseName of the node to find.

        Returns:
            int: The internal id of the node.
        """
        return self.__ua_nodeclass_by_browsename(browsename, "VariableType")

    def data_type_by_browsename(self, browsename: str) -> int:
        """Retrieves the internal id from the OPC UA graph which has a NodeClass "DataType"
        (UADataType in the UAGraph), for the given `browsename` input. If multiple nodes
        are found, an error is raised.

        Args:
            browsename (str): BrowseName of the node to find.

        Returns:
            int: The internal id of the node.
        """
        return self.__ua_nodeclass_by_browsename(browsename, "DataType")

    def object_by_browsename(self, browsename: str) -> int:
        """Retrieves the internal id from the OPC UA graph which has a NodeClass "Object"
        (UAObject in the UAGraph), for the given `browsename` input. If multiple nodes
        are found, an error is raised.

        Args:
            browsename (str): BrowseName of the node to find.

        Returns:
            int: The internal id of the node.
        """
        return self.__ua_nodeclass_by_browsename(browsename, "Object")

    def nodeid_by_browsename(
        self, browsename: str, nodeclass: Optional[str] = None
    ) -> int:
        """Retrieves the NodeId from the OPC UA graph which has a NodeClass `nodeclass`
        (UAVariableType in the UAGraph), for the given `browsename` input. If multiple nodes
        are found, an error is raised. Please note that the 'nodeclass' variable has to be
        a string, e.g. "VariableType", "DataType", etc. in the form of CamelCase.

        Args:
            browsename (str): BrowseName of the node to find.
            nodeclass (Optional[str], optional): NodeClass of the node to find. Defaults to None.

        Raises:
            ValueError: If browsename is None or empty string.
            ValueError: If no node is found.
            ValueError: If multiple nodes are found.

        Returns:
            int: The NodeId of the node.
        """
        if nodeclass:
            id = self.__ua_nodeclass_by_browsename(browsename, nodeclass)
        else:
            id = self.__ua_nodeclass_by_browsename(browsename)

        if id is None:
            raise ValueError(f"Could not find node with browsename {browsename}")
        node_id = self.nodes.loc[self.nodes["id"] == id, "NodeId"].values[0]
        return node_id

    def write_nodeset(
        self,
        filename_or_stringio: Union[str, StringIO],
        namespace_uri: str,
        include_outgoing_instance_level_references: Optional[bool] = True,
        last_modified: Optional[datetime] = None,
        publication_date: Optional[datetime] = None,
    ):
        if namespace_uri not in self.namespaces:
            raise ValueError("Could not find namespace uri: " + namespace_uri)
        namespace_index = self.namespaces.index(namespace_uri)
        if not include_outgoing_instance_level_references:
            use_references = self.remove_instance_level_outgoing_references(
                namespace_index=namespace_index
            )
        else:
            use_references = self.references

        # Always serialize namespace 1 in xml, so we need to remap indices
        new_namespaces_list = [self.namespaces[0], self.namespaces[namespace_index]]
        for i, n in enumerate(self.namespaces):
            if i != 0 and i != namespace_index:
                new_namespaces_list.append(n)

        remapper = {
            i: new_namespaces_list.index(n) for i, n in enumerate(self.namespaces)
        }
        use_nodes = self.nodes.copy()
        use_nodes["BrowseNameNamespace"] = use_nodes["BrowseNameNamespace"].astype(
            pd.Int32Dtype()
        )
        use_nodes["NodeId"] = use_nodes["NodeId"].map(
            lambda x: UANodeId(
                namespace=remapper[x.namespace],
                value=x.value,
                nodeid_type=x.nodeid_type,
            )
        )
        use_nodes["BrowseNameNamespace"] = use_nodes["BrowseNameNamespace"].map(
            remapper
        )

        create_nodeset2_file(
            nodes=use_nodes,
            references=use_references,
            models=self.models,
            serialize_namespace=1,
            namespaces=new_namespaces_list,
            filename_or_stringio=filename_or_stringio,
            last_modified=last_modified,
            publication_date=publication_date,
        )

    def get_normalized_nodes_df(self, namespace_uri: Optional[str] = None):
        lookup_df = create_lookup_df(self.nodes)
        if namespace_uri is not None:
            ns = self.namespaces.index(namespace_uri)
            nodes = self.nodes[self.nodes["ns"] == ns].copy()
        else:
            nodes = self.nodes.copy()
        nodes = denormalize_nodes_nodeids(nodes, lookup_df)
        nodes = nodes.drop(columns=["id"])
        nodes = nodes.sort_values(by=nodes.columns.values.tolist(), ignore_index=True)
        return nodes

    def __get_references_df(self, namespace_uri: str):
        ns = self.namespaces.index(namespace_uri)
        nodes_ns = self.nodes.loc[self.nodes["ns"] == ns, ["id"]].set_index("id")
        nodes_ns["in_ns"] = True
        references = (
            self.references.set_index("Src", drop=False)
            .join(nodes_ns)
            .rename(columns={"in_ns": "src_in_ns"}, errors="raise")
        )
        references = (
            references.set_index("Trg", drop=False)
            .join(nodes_ns)
            .rename(columns={"in_ns": "trg_in_ns"}, errors="raise")
        )
        references = references.loc[
            (references["src_in_ns"] == True) | (references["trg_in_ns"] == True),
            ["Src", "Trg", "ReferenceType"],
        ].copy()
        return references

    def get_normalized_references_df(self, namespace_uri: Optional[str] = None):
        lookup_df = create_lookup_df(self.nodes)
        if namespace_uri is not None:
            references = self.__get_references_df(namespace_uri)
        else:
            references = self.references.copy()

        references = denormalize_references_nodeids(references, lookup_df)
        references = references.sort_values(
            by=references.columns.values.tolist(), ignore_index=True
        )
        return references

    def remove_instance_level_outgoing_references(
        self, namespace_index: int
    ) -> pd.DataFrame:
        hmr = self.reference_type_by_browsename("HasModellingRule")
        htd = self.reference_type_by_browsename("HasTypeDefinition")
        self.nodes["ns"] = self.nodes["NodeId"].map(lambda x: x.namespace)
        ids_in_ns = self.nodes.loc[self.nodes["ns"] == namespace_index, ["id"]].copy()
        ids_in_ns = ids_in_ns.set_index("id", drop=False)

        references = self.references.set_index("Trg", drop=False)

        references = references[
            references.index.isin(ids_in_ns.index)
            | (references["ReferenceType"] == hmr)
            | (references["ReferenceType"] == htd)
        ].copy()

        return references

    def get_browsenames_for_nodeclass(
        self, node_class: str, namespace: Optional[int] = None
    ) -> List[str]:
        """This function will provide the option of returning the list of references
        present in the graph. The number of the namespace in which the reference type
        is defined can be specified to further narrow the search. If none is provided
        it will return all namespaces"""

        object_type_nodes = self.nodes[self.nodes["NodeClass"] == node_class]

        if namespace is not None:
            object_type_nodes.reset_index(inplace=True)
            mask = object_type_nodes["NodeId"].map(lambda x: x.namespace) == namespace
            object_type_nodes = object_type_nodes[mask]

        return object_type_nodes["BrowseName"].unique().tolist()

    def _get_browsename_from_id(self, id: int) -> str:
        row = self.nodes[self.nodes["id"] == id]
        return row["BrowseName"].values[0]

    def get_nodes_classes(self):
        return self.nodes["NodeClass"].unique().tolist()

    def get_instances_with_type_info(self):
        """Returns the UAObjects and the browsename of their type"""
        nodes = self.nodes[["NodeClass", "BrowseName", "NodeId", "id"]].set_index(
            "id", drop=False
        )
        has_type_def_id = self.reference_type_by_browsename("HasTypeDefinition")
        references_with_htd_id = self.references.loc[
            self.references["ReferenceType"] == has_type_def_id
        ]
        references_with_htd_id = references_with_htd_id[["Src", "Trg"]]
        typeinfo = nodes.rename(
            columns={
                "BrowseName": "TypeBrowseName",
                "NodeId": "TypeNodeId",
                "id": "Typeid",
            },
            errors="raise",
        ).drop(columns=["NodeClass"])
        references_with_nodes = (
            references_with_htd_id.set_index("Src")
            .join(nodes)
            .set_index("Trg")
            .join(typeinfo)
        )
        return references_with_nodes

    def get_enum_dict(self, enum_name: str):
        """This function will return the enum given its name in
        the form of a dict."""

        enum_node = self.nodes[
            (self.nodes.NodeClass == "UADataType")
            & (self.nodes.BrowseName == enum_name)
        ]

        if enum_node.shape[0] == 0:
            raise ValueError("The enum was not found in the graph")

        if enum_node.shape[0] > 1:
            raise ValueError("The enum was found multiple times in the graph")

        has_property_id = self.reference_type_by_browsename("HasProperty")
        node_id = enum_node["id"].values[0]
        outgoing_reference_row = self.references[
            (self.references.ReferenceType == has_property_id)
            & (self.references.Src == node_id)
        ]
        outgoing_id = outgoing_reference_row["Trg"].values[0]
        has_property_node = self.nodes[self.nodes["id"] == outgoing_id]
        has_property_name = has_property_node["BrowseName"].values[0]
        enumeration_datatypes = ["EnumStrings", "EnumValues"]
        if has_property_name not in enumeration_datatypes:
            raise ValueError(
                "The ReferenceType associated with the enum_name is not an EnumStrings or EnumValues"
            )

        enum_dict = dict()

        if has_property_name == "EnumStrings":
            ua_list_of = has_property_node["Value"].values[0]
            for localized_text_i, localized_text in enumerate(ua_list_of.value):
                # For case insensitivity captializing first word if string
                if isinstance(localized_text.text, str):
                    enum_dict[localized_text_i] = localized_text.text.title()
                else:
                    enum_dict[localized_text_i] = localized_text.text
        else:
            raise NotImplementedError("EnumValues not implemented")

        return enum_dict

    def get_enum_string(self, enum_name: str, number: int):
        """This function will return the string value of within an enum,
        provided you get both the of the enum and an integer value."""
        enum_dict = self.get_enum_dict(enum_name)
        # TODO: Cache enumdict for performance
        return enum_dict[number]

    def get_enum_int(self, enum_name: str, string: str):
        """This function will return the integer value of within an enum,
        provided you get both the of the enum and a string value."""
        # Ad Hoc solution to ensure text in files does not include newlines and line breaks
        string = string.replace("\r", "")
        string = string.replace("\n", "")
        string = string.title()
        string = string.strip()
        enum_dict = self.get_enum_dict(enum_name)
        # TODO: Cache enumdict for performance
        # Getting the dict key based on the value
        key_list = list(enum_dict.keys())
        val_list = list(enum_dict.values())
        try:
            position = val_list.index(string)
            return key_list[position]
        except Exception as get_enum_int_error:
            raise ValueError(
                "Could not find the string: {}, for the enum_name: {}".format(
                    string, enum_name
                )
            ) from get_enum_int_error

    def get_objects_of_type(self, type_name: str):
        has_type_def = self.reference_type_by_browsename("HasTypeDefinition")
        object_type_df = self.object_type_by_browsename(type_name)
        reference_ids = self.references[
            (self.references["ReferenceType"] == has_type_def)
            & (self.references["Trg"] == object_type_df)
        ]
        return self.nodes.loc[self.nodes["id"].isin(reference_ids["Src"])]

    def get_neighboring_nodes_by_browsename(
        self, browse_name: str, node_type: str, relation: str
    ):
        """This function will returning the references either pointing to
        or from a node. The results contain the source and target contents of
        the `references` table with actual text 'BrowseNames' and the corresponding
        node 'NodeIdes'. The 'relation' parameter is special since it must
        either be "outgoing" or "incoming". When inputting "child" it will return
        the outgoing references of the node, while "parent" will provide the
        references which point to the specific node."""

        id = None
        if node_type == "UAObject":
            id = self.object_by_browsename(browse_name)
        elif node_type == "UADataType":
            id = self.data_type_by_browsename(browse_name)
        elif node_type == "UAReferenceType":
            id = self.reference_type_by_browsename(browse_name)
        elif node_type == "UAObjectType":
            id = self.object_type_by_browsename(browse_name)
        elif node_type == "UAVariableType":
            id = self.variable_type_by_browsename(browse_name)
        else:
            raise ValueError(
                f"node_type, {node_type}, is not one of the proper types: "
                f"UAObject, UADataType, UAReferenceType, UAObjectType, UAVariableType"
            )

        if not id:
            raise ValueError(f"The id was not properly set and is {id}")

        return self.get_neighboring_nodes_by_id(id, relation)

    def get_neighboring_nodes_by_id(self, id: int, relation: str):

        direction = None
        if relation == "outgoing":
            direction = "Src"
        elif relation == "incoming":
            direction = "Trg"

        if direction == None:
            raise ValueError(
                "The relation of the neighbouring nodes must be"
                + 'indicated by setting "relation" to either "outgoing" or "incoming".'
            )

        # Creating duplicate for safety's sake
        references = self.references.copy()

        # Getting getting the subset of references which only contain the desired node
        references = references[references[direction] == id]

        # Creating a dataframe with only names to conduct joins and get names
        names = self.nodes[["BrowseName", "NodeId", "id"]].set_index("id", drop=True)

        # Getting the names of the ReferenceTypes into the tables instead of numbers
        references.set_index("ReferenceType", drop=True, inplace=True)
        references = references.join(names, how="inner")
        references.rename(columns={"BrowseName": "ReferenceType"}, inplace=True)
        references.drop("NodeId", axis=1, inplace=True)

        # Getting names of the ids in the Src Column, joining in the same manner as before
        references.set_index("Src", drop=False, inplace=True)
        references = references.join(names, how="inner")
        references.rename(
            columns={"BrowseName": "Source", "NodeId": "SourceNodeId"}, inplace=True
        )

        # Getting names of the ids in the Trg Column, joining in the same manner as before
        references.set_index("Trg", drop=False, inplace=True)
        references = references.join(names, how="inner")
        references.rename(
            columns={"BrowseName": "Target", "NodeId": "TargetNodeId"}, inplace=True
        )

        return references

    def create_node_paths_by_reference_types(
        self,
        root_node_browsename: str,
        hierarchical_reference_types: List[str],
    ) -> pd.DataFrame:
        root_node_id = self.object_by_browsename(root_node_browsename)

        reference_type_ids = []
        for ref in hierarchical_reference_types:
            try:
                reference_type_ids.append(self.reference_type_by_browsename(ref))
            except IndexError as e:
                logger.error(
                    f"The ReferenceType: {ref}, was not found in the nodes and references"
                )
                raise

        hierarchical_references = self.references[
            self.references["ReferenceType"].isin(reference_type_ids)
        ]
        root_node_id_df = self.nodes.loc[
            self.nodes["id"] == root_node_id, ["id"]
        ].copy()
        relatives = find_relatives(
            nodes=root_node_id_df,
            nodes_key_col="id",
            edges=hierarchical_references,
            relative_type="d",
            keep_paths=True,
        )

        path_columns = [c for c in relatives.columns if type(c) == int and c > 0]

        relatives = (
            relatives.drop("id", axis=1)
            .melt(id_vars=[0, "end"], value_vars=path_columns, value_name="id")
            .dropna()
        )
        nodes_browsename = self.nodes[["id", "BrowseName"]].set_index("id")

        relatives = (
            relatives.set_index("id").join(nodes_browsename).reset_index(drop=True)
        )
        relatives = relatives.sort_values(by=[0, "end", "variable"], ignore_index=True)
        relatives = relatives.rename(columns={0: "start"})
        relatives = (
            relatives.groupby(by=["start", "end"])
            .agg({"BrowseName": lambda x: "/".join(x.to_list())})
            .reset_index()
        )

        node_paths = relatives[["end", "BrowseName"]].rename(
            columns={"end": "id", "BrowseName": "NodePath"}
        )
        root_node_id_df_to_add = pd.DataFrame({"id": [root_node_id], "NodePath": [""]})
        node_paths = pd.concat([node_paths, root_node_id_df_to_add])

        # Adding the root_node to the start of the path
        node_paths["NodePath"] = node_paths["NodePath"].apply(
            lambda name: "/".join([root_node_browsename, name])
        )
        node_paths = node_paths.sort_values("NodePath").reset_index(drop=True)

        return node_paths

    def find_circular_reference_nodes(self, namespace_uri: str) -> pd.DataFrame:
        """This function finds circular references in the given namespace. Returns a dataframe with the Node IDs involved in the circular references"""
        lookup_df = create_lookup_df(self.nodes)
        refs_ns = self.__get_references_df(namespace_uri)
        type_references = self.references
        type_nodes = self.nodes
        hierarchy_refs = hierarchical_references(refs_ns, type_references, type_nodes)
        transitive_closures = fast_transitive_closure(hierarchy_refs)
        cyclic_refs = pd.merge(
            transitive_closures,
            transitive_closures,
            how="inner",
            left_on=["Src", "Trg"],
            right_on=["Trg", "Src"],
        )
        cyclic_refs = pd.DataFrame(cyclic_refs["Src_x"].unique(), columns=["Src"])
        refcols = ["Src"]
        for col in refcols:
            uniques = lookup_df.rename(columns={"uniques": col}, errors="raise")
            cyclic_refs = (
                cyclic_refs.set_index(col).join(uniques).reset_index(drop=True)
            )
        cyclic_refs = cyclic_refs.rename(columns={"Src": "NodeId"})
        return cyclic_refs

    def __get_nodes_from_ns(self, namespace_uri: str):
        ns = self.namespaces.index(namespace_uri)
        nodes_ns = self.nodes.loc[self.nodes["ns"] == ns, ["id"]].set_index("id")
        return nodes_ns
